/* Copyright 1991 Digital Equipment Corporation.               */
/* All rights reserved. See copyright notice in attached file COPYRIGHT. */
/* Last modified on Mon Aug 19 22:33:25 1991 by burrows        */
/* an implementation of malloc, free and realloc, that is wasteful of time
 * and header space, but has little internal fragmentation.
 *  Should be used only if you have a few large objects.
 */
#include <stdio.h>

#define abort myabort

#define MAGIC1   0x12345435
#define MAGIC2   0x21434523
#define ALIGN	(sizeof (char *))
#define SMALLEST	(sizeof (struct area))
#define CLICK	(32*1024)

extern char *sbrk ();
extern void abort ();
extern void bcopy ();

static unsigned round (x, y)
unsigned x;
unsigned y;
{
	x += y-1;
	x -= x % y;

	return (x);
}

struct area {
	unsigned magic1;
	char *file;
	int line;
	struct area *usenext;
	struct area *uselast;
	struct area *next;
	struct area *last;
	unsigned size;
	unsigned magic2;
};

static struct area head = { MAGIC1, "", 0, 0, 0, &head, &head, 0, MAGIC2 };
static struct area usehead = { MAGIC1, "", 0, &usehead, &usehead, 0, 0, 0, MAGIC2 };

char *mymalloc (n, file, line)
register unsigned n;
char *file;
int line;
{
	register struct area *a;
	register struct area *h = &head;
	register struct area *best;
	char *p;
	unsigned newsize;

	check (1);

	n = round (n, ALIGN);
	if (n < SMALLEST)
		n = SMALLEST;

	best = 0;
	a = h->next;
	while (a != h) {
		if (a->size >= n) {
			if (best == 0 || a->size < best->size)
				best = a;
		}
		a = a->next;
	}

	if (best == 0) {
		newsize = round (n + sizeof (struct area), CLICK);
		p = sbrk ((int) newsize);

		if (p == (char *) -1) {
			return (0);
		}

		best = (struct area *) p;
		best->magic1 = MAGIC1;
		best->next = h;
		best->last = h->last;
		best->last->next = best->next->last = best;
		best->size = newsize - sizeof (struct area);
		best->file = "";
		best->line = 0;
		best->magic2 = MAGIC2;
	} else if (best->magic1 != MAGIC1 || best->magic2 != MAGIC2) {
		(void) fprintf (stderr, "malloc: bad magic number %x\n", best);
		status ();
		abort ();
	}

	if (best->size >= n + sizeof (struct area) + SMALLEST) {
		a = (struct area *) (((char *)best) + sizeof (struct area) + n);
		a->magic1 = MAGIC1;
		a->next = best->next;
		a->last = best;
		a->last->next = a->next->last = a;
		a->size = best->size - n - sizeof (struct area);
		a->magic2 = MAGIC2;

		best->size = n;
	}

	best->next->last = best->last;
	best->last->next = best->next;
	best->next = (struct area *) MAGIC1;
	best->last = (struct area *) MAGIC2;

	best->usenext = usehead.usenext;
	best->uselast = &usehead;
	best->usenext->uselast = best;
	best->uselast->usenext = best;

	best->file = file;
	best->line = line;

	check (2);

	return (((char *) best) + sizeof (struct area));
}

void free (p)
char *p;
{
	register struct area *h = &head;
	register struct area *a;
	register struct area *freed;

	check (3);

	if (p == 0) {
		return;
	}

	freed = (struct area *) (p - sizeof (struct area));

	if (freed->magic1 != MAGIC1 || freed->magic2 != MAGIC2) {
		(void) fprintf (stderr, "free: bad magic number (1) %x\n", p);
		status ();
		abort ();
	}
	if (freed->next != (struct area *) MAGIC1 ||
	    freed->last != (struct area *) MAGIC2) {
		(void) fprintf (stderr, "free: bad magic number (2) %x\n", p);
		status ();
		abort ();
	}
	freed->usenext->uselast = freed->uselast;
	freed->uselast->usenext = freed->usenext;
	a = h->next;
	while (a != h && a < freed)
		a = a->next;

	if (a->last != h &&
	    (((char *) a->last) + sizeof (struct area) + a->last->size) ==
			(char *) freed) {
		if (a != h &&
		    (((char *) freed) + sizeof (struct area) + freed->size) ==
			(char *) a) {
			a->last->size += freed->size + sizeof (struct area) +
					 a->size + sizeof (struct area);
			a->last->next = a->next;
			a->next->last = a->last;
			freed->magic1 = freed->magic2 = 0;
			freed->next = freed->last = 0;
			a->magic1 = a->magic2 = 0;
			a->next = a->last = 0;
		} else {
			a->last->size += freed->size + sizeof (struct area);
			freed->magic1 = freed->magic2 = 0;
			freed->next = freed->last = 0;
		}
	} else if (a != h &&
		   (((char *) freed) + sizeof (struct area) + freed->size) ==
			(char *) a) {
		freed->next = a->next;
		freed->last = a->last;
		freed->last->next = freed->next->last = freed;
		freed->size += a->size + sizeof (struct area);
		a->magic1 = a->magic2 = 0;
		a->next = a->last = 0;
	} else {
		freed->next = a;
		freed->last = a->last;
		freed->last->next = freed->next->last = freed;
	}

	check (4);
}

char *myrealloc (p, n, file, line)
char *p;
unsigned n;
char *file;
int line;
{
	register struct area *h = &head;
	register struct area *a;
	register struct area *b;
	register struct area *freed;
	char *newp;
	unsigned newsize;

	check (5);

	if (p == 0) {
		return (mymalloc (n, file, line));
	}
	

	n = round (n, ALIGN);
	if (n < SMALLEST)
		n = SMALLEST;

	freed = (struct area *) (p - sizeof (struct area));

	if (freed->magic1 != MAGIC1 || freed->magic2 != MAGIC2) {
		(void) fprintf (stderr,"realloc: bad magic number (1) %x\n", p);
		status ();
		abort ();
	}
	if (freed->next != (struct area *) MAGIC1 ||
	    freed->last != (struct area *) MAGIC2) {
		(void) fprintf (stderr,"realloc: bad magic number (2) %x\n", p);
		status ();
		abort ();
	}

	a = h->next;
	while (a != h && a < freed)
		a = a->next;

	if (n > freed->size && a == h) {
		newsize = round (n - freed->size , CLICK);
		newp = sbrk ((int) newsize);

		if (newp != (char *) -1) {
			a = (struct area *) newp;
			a->magic1 = MAGIC1;
			a->next = h;
			a->last = h->last;
			a->last->next = a->next->last = a;
			a->size = newsize - sizeof (struct area);
			a->file = "";
			a->line = 0;
			a->magic2 = MAGIC2;
		}
	}

	if (n > freed->size && a != h &&
	    (((char *) freed) + freed->size) == (char *) a &&
	    n <= freed->size + a->size + sizeof (struct area)) {
		a->last->next = a->next;
		a->next->last = a->last;
		a->magic1 = a->magic2 = 0;
		a->next = a->last = 0;
		freed->size += a->size + sizeof (struct area);
	}

	if (n <= freed->size) {
		if (freed->size >= n + sizeof (struct area) + SMALLEST) {
			b = (struct area *) (((char *) freed) +
				sizeof (struct area) + n);
			b->magic1 = MAGIC1;
			b->next = a;
			b->last = a->last;;
			b->last->next = b->next->last = b;
			b->size = freed->size - n - sizeof (struct area);
			b->magic2 = MAGIC2;

			freed->size = n;
		}

		check (6);

		return (((char *) freed) + sizeof (struct area));
	}

	newp = mymalloc (n, file, line);

	if (newp != 0) {
		bcopy (p, newp, (int) freed->size);
		free (p);
	}

	return (newp);
}

check (i)
int i;
{
	struct area *x;

	for (x = usehead.usenext; x != &usehead; x = x->usenext)
		if (x->magic1 != MAGIC1 || x->magic2 != MAGIC2) {
			(void) fprintf (stderr, "malloc check failed 1 %d\n",i);
			status ();
			abort ();
		}
	for (x = head.next; x != &head; x = x->next)
		if (x->magic1 != MAGIC1 || x->magic2 != MAGIC2) {
			(void) fprintf (stderr, "malloc check failed 2 %d\n",i);
			status ();
			abort ();
		}
}

status ()
{
	struct area *x;

	(void) setuid (getuid ());

	for (x = usehead.usenext; x != &usehead; x = x->usenext)
		(void) fprintf (stderr, "inuse %x %x %s %d %x %x %x %x %d %x\n",
			x, x->magic1, x->file, x->line,
			x->usenext, x->uselast, x->next, x->last,
			x->size, x->magic2);
	for (x = head.next; x != &head; x = x->next)
		(void) fprintf (stderr, "free %x %x %s %d %x %x %x %x %d %x\n",
			x, x->magic1, x->file, x->line,
			x->usenext, x->uselast, x->next, x->last,
			x->size, x->magic2);
}

char *malloc (n)
unsigned n;
{
	return (mymalloc (n, "malloc stub", 0));
}

char *realloc (p, n)
char *p;
unsigned n;
{
	return (myrealloc (p, n, "malloc stub", 0));
}
