/* Copyright 1991 Digital Equipment Corporation.               */
/* All rights reserved. See copyright notice in attached file COPYRIGHT. */
/* Last modified on Mon Aug 19 22:33:14 1991 by burrows        */
#include <stdio.h>
#include <errno.h>

extern void perror ();
extern void exit ();
extern long atol ();
extern char *realloc ();

static void delete ();

main (argc, argv)
int argc;
char *argv[];
{
	register long *table = 0;
	register long max = 0;
	register long i, n;
	FILE *fp;
	char buf[4096];
	long newmax;

	if (argc != 2) {
		(void) fprintf (stderr, "usage: msgdelete deletefile\n");
		exit (2);
	}

	if ((fp = fopen (argv[1], "r")) == 0) {
		perror (argv[1]);
		exit (2);
	}

	while (fgets (buf, sizeof (buf), fp) != 0) {
		i = atol (buf);
		if (i < 0) {
			(void) fprintf (stderr,
			    "msgdelete: negative integer in delete file\n");
			exit (2);
		}
		if (i >= max) {
			newmax = i + 1024;
			table = (long *) realloc ((char *) table,
				(unsigned) newmax * sizeof (long));
			if (table == 0) {
				perror ("realloc");
				exit (2);
			}
			bzero ((char *) &table[max],
				(newmax-max) * sizeof (long));
			max = newmax;
		}
		table[i] = -1;
	}

	if (ferror (fp)) {
		perror (argv[1]);
		exit (2);
	}
	(void) fclose (fp);

	for (i = n = 0; i < max; i++) {
		if (table[i] >= 0)
			table[i] = n++;
	}
	n = max - n;

	delete (n, max, table);

	if (ferror (stdin)) {
		perror ("stdin");
		exit (2);
	}

	(void) fflush (stdout);
	if (ferror (stdout) || (fsync (fileno (stdout)) < 0 && errno!=EINVAL)) {
		perror ("stdout");
		exit (2);
	}

	exit (0);
}

static void delete (diff, max, table)
long diff, max;
long *table;
{
	char bufin[2000];
	char bufout[2000];
	register int c;
	register char *s, *p;
	register long n, ln, oln;
	int shift;

	c = getchar ();
	bufout[0] = '\0';

	while ((c & 0xc0) == 0x80) {
		c &= 0x3f;
		s = &bufin[c];

		c = getchar ();
		
		while (c != '\0' && c != EOF) {
			*s++ = c;
			c = getchar ();
		}
		if (c != EOF)
			c = getchar ();

		*s = '\0';

		oln = -1;
		ln = 0;

		while ((c & 0xc0) != 0x80 && c != EOF) { 
			n = 0;
			shift = 0;
			while ((c & 0xc0) == 0xc0 && c != EOF) {
				n |= (c & 0x3f) << shift;
				shift += 6;
				c = getchar ();
			}
			n |= (c & 0x7f) << shift;
			n += ln;
			ln = n;
			if (n >= max)
				n -= diff;
			else if ((n = table[n]) < 0) {
				c = getchar ();
				continue;
			}

			if (oln == -1) {
				p = bufin;
				s = bufout;
				while (*s != '\0' && *p == *s)
					s++, p++;

				if (s-bufout > 0x3f) {
					(void) putchar (0x80 | 0x3f);
					(void) fwrite (bufout+0x3f, 1,
						(s-bufout)-0x3f, stdout);
				} else
					(void) putchar (0x80 | (s-bufout));

				while (*p != '\0')
					(void) putchar (*s++ = *p++);
				*s = '\0';
				(void) putchar ('\0');

				oln = n;
			} else {
				n -= oln;
				oln += n;
			}

			while (n > 0x7f) {
				(void) putchar (0xc0 | n);
				n >>= 6;
			}
			(void) putchar (n);
			c = getchar ();
		}
	}
}
